"""
Python module to interface with wrapped meshfix

"""
import numpy as np
import ctypes
import warnings
from pymeshfix import _meshfix

try:
    import vtkInterface as vtki
    vtkenabled = True
except BaseException:
    warnings.warn('Unable to import vtkInterface.  Will be unable to plot meshes')
    vtkenabled = False


class MeshFix(object):
    """
    Cleans and tetrahedralize surface meshes using MeshFix

    Parameters
    ----------
    args : vtkInterface.PolyData or (np.ndarray, np.ndarray)
        Either a vtkInterface surface mesh or a nx3 vertex array and nx3 face
        array.

    """

    def __init__(self, *args):
        """ initializes MeshFix """
        if isinstance(args[0], vtki.PolyData):
            mesh = args[0]
            self.v = mesh.points
            self.f = mesh.GetNumpyFaces(force_C_CONTIGUOUS=True)
        elif isinstance(args[0], np.ndarray):
            self.LoadArrays(args[0], args[1])
        else:
            raise Exception('Invalid input')

    def LoadArrays(self, v, f):
        """
        Loads triangular mesh from vertex and face arrays

        Face arrays/lists are v and f.  Both vertex and face arrays should be
        2D arrays with each vertex containing XYZ data and each face containing
        three points
        """
        # Check inputs
        if not isinstance(v, np.ndarray):
            try:
                v = np.asarray(v, np.float)
                if v.ndim != 2 and v.shape[1] != 3:
                    raise Exception(
                        'Invalid vertex format.  Shape should be (npoints, 3)')
            except BaseException:
                raise Exception(
                    'Unable to convert vertex input to valid numpy array')

        if not isinstance(f, np.ndarray):
            try:
                f = np.asarray(f, ctypes.c_int)
                if f.ndim != 2 and f.shape[1] != 3:
                    raise Exception(
                        'Invalid face format.  Shape should be (nfaces, 3)')
            except BaseException:
                raise Exception(
                    'Unable to convert face input to valid numpy array')

        # Store to self
        self.v = v
        self.f = f

    @property
    def mesh(self):
        """ the surface mesh """
        if vtkenabled:
            return vtki.MeshfromVF(self.v, self.f)
        else:
            raise Exception('Cannot generate mesh without vtkInterface.\n' +
                            'Run: pip install vtkInterface')

    def Plot(self, showbound=True):
        """
        Plot the mesh.

        Parameters
        ----------
        showbound : bool, optional
            Shows boundaries.  Default True
        """
        if showbound:
            vtki.PlotBoundaries(self.mesh, showedges=True)
        else:
            self.mesh.Plot(showedges=True)

    def Repair(self, verbose=False, joincomp=False, removeSmallestComponents=True):
        """
        Performs mesh repair using MeshFix's default repair process

        Parameters
        ----------
        verbose : bool, optional
            Enables or disables debug printing.  Disabled by default.

        joincomp : bool, optional
            Attempts to join nearby open components.
        
        removeSmallestComponents : bool, optional
            Remove all but the largest isolated component from the mesh
            before beginning the repair process.  Default True

        Notes
        -----
        Vertex and face arrays are updated inplace.  Access them with:
        meshfix.v
        meshfix.f

        """
        self.v, self.f = _meshfix.CleanFromVF(self.v, self.f, verbose,
                                              joincomp, removeSmallestComponents)

    def Write(self, filename, ftype=None, binary=True):
        """
        Writes a surface mesh to disk.

        Written file may be an ASCII or binary ply, stl, or vtk mesh file.

        Parameters
        ----------
        filename : str
            Filename of mesh to be written.  Filetype is inferred from the
            extension of the filename unless overridden with ftype.  Can be
            one of the following types (.ply, .stl, .vtk)

        ftype : str, optional
            Filetype.  Inferred from filename unless specified with a three
            character string.  Can be one of the following: 'ply',  'stl', or
            'vtk'.

        Notes
        -----
        Binary files write much faster than ASCII.
        """
        if not vtkenabled:
            raise Exception('Cannot save mesh without vtk.  ' +
                            'Please install vtk with python bindings.')

        self.mesh.Write()



"""
Python/Cython wrapper for MeshFix by Marco Attene
"""
import ctypes
import warnings

from libcpp cimport bool

import numpy as np
cimport numpy as np

ctypedef unsigned short UINT16

""" Wrapped tetgen class """
cdef extern from "meshfix.h" namespace "T_MESH":
    cdef cppclass Basic_TMesh_wrap:
        # Constructor
        Basic_TMesh_wrap()

        # Inputs     
        int load(const char*)
        int loadArray(int, double*, int, int*)
        
        # MeshFix
        int removeSmallestComponents()
        int boundaries()
        int fillSmallBoundaries(int, bool)
        int meshclean(int, int)
        int save(const char*, bool)

        # Select self intersecting triangles
        int selectIntersectingTriangles(UINT16, bool)
        void GetSelected(int*)
#        void IntersectionRemoval()

        # Settings
        void SetVerbose(int)
        
        # Outputs
        int ReturnTotalPoints()
        int ReturnTotalFaces()
        void PopArrays(double*, int*)

        # MISC        
        void Join()
        

cdef class PyTMesh:
    """
    Python class to interface with C++ Basic_TMesh object

    MeshFix V2.0 - by Marco Attene
    If MeshFix is used for research purposes, please cite the following paper:

    M. Attene.\n   A lightweight approach to repairing digitized polygon meshes
    The Visual Computer, 2010. (c) Springer.
    
    """
    
    cdef Basic_TMesh_wrap c_tmesh      # hold a C++ instance which we're wrapping
    
    def __cinit__(self, quiet=1):
        """ Create TMesh object """
        self.c_tmesh = Basic_TMesh_wrap()
        
        # Enable/Disable printed progress
        self.c_tmesh.SetVerbose(not quiet)

    def LoadFile(self, filename):
        """
        Loads mesh from file
        
        Currently, the following file formats are supported:
        Open Inventor (IV), VRML 1.0 and 2.0 (WRL), Object File Format (OFF),
        IMATI Ver-Tri (VER, TRI), PLY, OBJ, STL.
        
        The loader automatically reconstructs a manifold triangle connectivity
        """
        
        # Initializes triangulation
        
        #cdef char *cstring = _AsString(filename, "utf-8", "Error ~")
        py_byte_string = filename.encode('UTF-8')
        cdef char* cstring = py_byte_string
        result = self.c_tmesh.load(cstring);
        if result:
            raise IOError('MeshFix is unable to open {:s}'.format(filename))

    def SaveFile(self, filename, back_approx=False):
        """
        Saves cleaned mesh to file
                
        The file format is deduced from one of the following filename
        extensions:
            wrl = vrml 1.0
            iv = OpenInventor
            off = Object file format
            ply = PLY format
            tri = IMATI Ver-Tri
            
        If 'back_approx' is set to True, vertex coordinates are approximated
        to reflect the limited precision of floating point
        representation in ASCII files. This should be used when
        coherence is necessary between in-memory and saved data.
        A non-zero return value is returned if errors occur.
        """
        
        # Convert filename to c string and save
        py_byte_string = filename.encode('UTF-8')
        cdef char* cstring = py_byte_string
        #cdef char *cstring = _AsString(filename, "utf-8", "Error ~")
        result = self.c_tmesh.save(cstring, back_approx)

        if result:
            raise IOError('MeshFix is unable to save mesh to {:s}'.format(filename))

    def LoadArray(self, v, f):
        """
        Loads points from numpy verticies and faces arrays
        The loader automatically reconstructs a manifold triangle connectivity
        """
        if not v.flags['C_CONTIGUOUS']:
            if v.dtype != np.float:
                v = np.ascontiguousarray(v, dtype=np.float)
            else:
                v = np.ascontiguousarray(v)
    
        elif v.dtype != np.float:
            v = v.astype(np.float)
    
        # Ensure inputs are of the right type
        if not f.flags['C_CONTIGUOUS']:
            if f.dtype != ctypes.c_int:
                f = np.ascontiguousarray(f, dtype=ctypes.c_int)
            else:
                f = np.ascontiguousarray(f)
    
        elif f.dtype != ctypes.c_int:
            f = f.astype(ctypes.c_int)
            
        cdef int nv = v.shape[0]
        cdef double [::1] points = v.ravel()
    
        cdef int nt = f.shape[0]
        cdef int [::1] faces = f.ravel()
        
        # Load to C object
        self.c_tmesh.loadArray(nv, &points[0], nt, &faces[0])

    def MeshClean(self, int max_iters=10, inner_loops=3):
        """
        Iteratively call strongDegeneracyRemoval and strongIntersectionRemoval
        to produce an eventually clean mesh without degeneracies and intersections.
        The two aforementioned methods are called up to max_iter times and
        each of them is called using 'inner_loops' as a parameter.
        Returns true only if the mesh could be completely cleaned.
        """
        
        self.c_tmesh.meshclean(max_iters, inner_loops)

    def Boundaries(self):
        """ Get the number of boundary loops of the triangle mesh """
        return self.c_tmesh.boundaries()

    def RemoveSmallestComponents(self):
        return self.c_tmesh.removeSmallestComponents()

    def FillSmallBoundaries(self, nbe=0, refine=True):
        """
        Fills all the holes having at least 'nbe' boundary edges. If 'refine'
        is true, adds inner vertices to reproduce the sampling density
        of the surroundings. Returns number of holes patched.
        If 'nbe' is 0 (default), all the holes are patched.
        """
        return self.c_tmesh.fillSmallBoundaries(nbe, refine)

    def ReturnArrays(self):
        """ Return numpy arrays from self object """
        
        # Size point array
        cdef int npoints = self.c_tmesh.ReturnTotalPoints()
        cdef double [::1] points = np.empty(npoints*3)

        # Size face array
        cdef int nfaces = self.c_tmesh.ReturnTotalFaces()
        cdef int [::1] faces = np.empty(nfaces*3, ctypes.c_int)

        # Populate points array
        self.c_tmesh.PopArrays(&points[0], &faces[0])

        v = np.asarray(points).reshape((-1, 3))
        f = np.asarray(faces).reshape((-1, 3))
        
        return v, f

    def JoinClosestComponents(self):
        """
        Attempts to join nearby open components.  Should be run before mesh
        repair
        """
        self.c_tmesh.Join()

    def SelectIntersectingTriangles(self, UINT16 tris_per_cell, bool justproper):
        """
        
        Selects all the triangles that unproperly intersect other parts of
        the mesh and return their number. The parameter 'tris_per_cell'
        determines the depth of the recursive space subdivision used to keep
        the complexity under a resonable threchold. The default value is safe
        in most cases.
        
        if 'justproper' is true, coincident edges and vertices are not regarded
        as intersections even if they are not common subsimplexes.
        
        """
        
        # Returns the number of intersecting triangles
        its = self.c_tmesh.selectIntersectingTriangles(tris_per_cell, justproper)
        
        # Create a face array and populate it with the intersecting faces
        cdef int [::1] faces = np.empty(its, ctypes.c_int)
        self.c_tmesh.GetSelected(&faces[0])
        
        return np.asarray(faces)

#    def IntersectionRemoval(self):
#        """ Remove intersections """
#        self.c_tmesh.IntersectionRemoval()
#        

def CleanFromFile(infile, outfile, verbose=True, joincomp=False):
    """ Performs default cleaning procedure on input file """

    # Create mesh object and load from file
    tin = PyTMesh(verbose)
    tin.LoadFile(infile)
    if joincomp:
        tin.JoinClosestComponents()
    
    Repair(tin, verbose, joincomp)
    
    # Save to file
    if verbose:
        print('Saving repaired mesh to {:s}'.format(outfile))
    tin.SaveFile(outfile)

def CleanFromVF(v, f, verbose=True, joincomp=False, removeSmallestComponents=True):
    """
    Performs default cleaning procedure on vertex and face arrays

    Returns cleaned vertex and face arrays
    
    """

    # Create mesh object and load from file
    tin = PyTMesh(verbose)
    tin.LoadArray(v, f)

    # repari and return vertex and face arrays
    Repair(tin, verbose, joincomp, removeSmallestComponents)
    return tin.ReturnArrays()

def Repair(tin, verbose, joincomp, removeSmallestComponents=True):
    """ Performs mesh repair using default cleaning procedure """

    # Keep only the largest component (i.e. with most triangles)
    if removeSmallestComponents:
        sc = tin.RemoveSmallestComponents()
        if sc and verbose:
            print('Removed {:d} small components'.format(sc))

    # join closest components
    if joincomp:
        tin.JoinClosestComponents()
    
    if tin.Boundaries():
        if verbose:
            print('Patching holes...')
        holespatched = tin.FillSmallBoundaries()
        if verbose:
            print('Patched {:d} holes'.format(holespatched))
    
    # Perform mesh cleaning
    if verbose:
        print('Fixing degeneracies and intersections')
        
    tin.Boundaries()
    result = tin.MeshClean()

    # Check boundaries again
    if tin.Boundaries():
        if verbose:
            print('Patching holes...')
        holespatched = tin.FillSmallBoundaries()
        if verbose:
            print('Patched {:d} holes'.format(holespatched))
    
        if verbose:
            print('Performing final check...')
        result = tin.MeshClean()

    if result:
        warnings.warn('MeshFix could not fix everything')



"""
Setup for pymeshfix
"""
from io import open as io_open
import sys
import os

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext

import numpy

# Define macros for cython
macros = []
extra_args = []

# Check if 64-bit
if sys.maxsize > 2**32:
    macros.append(('IS64BITPLATFORM', None))

# Check if Windows
if not os.name == 'nt':
    extra_args.append('-std=gnu++11')
    # extra_args.append('-std=c++11') # might have to use this instead

# Files for Meshfix and cython wrapper
files = []

# Get version from version info
__version__ = None
version_file = os.path.join(os.path.dirname(__file__), 'pymeshfix',
                            '_version.py')
with io_open(version_file, mode='r') as fd:
    exec(fd.read())


setup(
    name='pymeshfix',
    packages=['pymeshfix'],
    version=__version__,
    description='Repairs triangular meshes',
    long_description=open('README.rst').read(),
    author='Alex Kaszynski',
    author_email='akascap@gmail.com',
    license='MIT',
    classifiers=['Development Status :: 4 - Beta',
                 'Intended Audience :: Science/Research',
                 'License :: OSI Approved :: MIT License',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3.6'],
    url='https://github.com/akaszynski/pymeshfix',

    # Build cython modules
    cmdclass={'build_ext': build_ext},
    ext_modules=[Extension("pymeshfix._meshfix",
                           ['pymeshfix/cython/meshfix.cpp',
                            'pymeshfix/cython/tin.cpp',
                            'pymeshfix/cython/checkAndRepair.cpp',
                            'pymeshfix/cython/coordinates.cpp',
                            'pymeshfix/cython/detectIntersections.cpp',
                            'pymeshfix/cython/edge.cpp',
                            'pymeshfix/cython/graph.cpp',
                            'pymeshfix/cython/heap.cpp',
                            'pymeshfix/cython/holeFilling.cpp',
                            'pymeshfix/cython/io.cpp',
                            'pymeshfix/cython/jqsort.cpp',
                            'pymeshfix/cython/list.cpp',
                            'pymeshfix/cython/marchIntersections.cpp',
                            'pymeshfix/cython/matrix.cpp',
                            'pymeshfix/cython/orientation.c',
                            'pymeshfix/cython/point.cpp',
                            'pymeshfix/cython/tmesh.cpp',
                            'pymeshfix/cython/triangle.cpp',
                            'pymeshfix/cython/vertex.cpp',
                            'pymeshfix/cython/_meshfix.pyx'],
                           language='c++',
                           extra_compile_args=extra_args,
                           define_macros=macros)],

    keywords='meshfix',
    include_dirs=[numpy.get_include()],
    package_data={'pymeshfix.tests': ['StanfordBunny.ply']},
    install_requires=['numpy>1.9.3', 'vtkInterface']
)
